#include "Tlist.h"
#include "Twordlist.h"
#include "Tlookuplist.h"

#define OPCODE_L 1
#define OPCODE_H 2
#define OPCODE_F 3
#define OPCODE_O 4
#define HTML 0
#define TEXT 1
#define LISP 2

void about () {
   printf("Natural Language Analysis:  Proper Nouns and Pro Nouns. (v 1.0)\n"
	  "  Written by James Dean Palmer (Petersen CS 485; Summer 1998)\n\n"
	  "This program takes a file of natural language (english) and then performs\n"
	  "various tests on the pronouns and proper nouns in an attempt to establish\n"
	  "the most likely relationships between them.\n\n"
	  "Although most of the metrics between proper nouns and pro nouns are very\n"
	  "exact (word distances, etc), the hueristics that weight and utilize this\n"
	  "information are very non-exact, and have no guarantees for reliability.\n\n"
	  "My paper on this program explains the approach I have taken in more\n"
	  "depth - it can be found at:\n\n"
	  "  http://www.tiger-marmalade.com/~james/code/nlanguage.html\n\n"
	  "  analyze [ options ]\n\n"
	  "  Options:\n"
	  "  -f FILENAME\n"
	  "     Use FILENAME as the plain text to be analyzed.\n"
	  "     (Default: \"default.txt\")\n\n"
	  "  -l FILENAME\n"
	  "     Use FILENAME as the world knowledge lookup table.\n"
	  "     (Default: \"lookup.txt\")\n\n"
	  "  -h INTEGER\n"
	  "     Use huersitic number INTEGER to weight results. (0-2)\n"
	  "     (Default: 0)\n\n"
	  "  -o INTEGER\n"
	  "     INTEGER determines the output style.  HTML=0, Text=1, Lispish=2.\n"
	  "     (Default: 0)\n\n");
}

int main (int argc, char *argv[]) {
   Tlist * lookup_list;
   Tlist * word_list;
   Tlist * pronoun_list;
   Tword * current_word;
   Tword * current_pronoun;
   Tlookup * current_lookup;
   int i=0, j, opcode, hueristic, output;
   char a[20];
   char b[20];
   char *filename_word;
   char *filename_lookup;

   // Set the program defaults.
   filename_word = &a[0];
   filename_lookup = &b[0];
   strcpy (filename_word, "default.txt");
   strcpy (filename_lookup, "lookup.txt");
   output = 0;
   hueristic = 0;
   opcode = 0;

   // 0.  Accept arguments from the program arguments.

   for (j = 1; j < argc; j=j+1) {
      switch (opcode) {
       case 0:
	 if (strcmp(argv[j], "-l") == 0) opcode = OPCODE_L;       // Look up option.
	 else if (strcmp(argv[j], "-h") == 0) opcode = OPCODE_H;  // Hueristic.
	 else if (strcmp(argv[j], "-f") == 0) opcode = OPCODE_F;  // File name.
	 else if (strcmp(argv[j], "-o") == 0) opcode = OPCODE_O;
	 else {
	    printf("Illegal argument.\n\n");
	    about();
	    exit (8);
	 }
	 break;
       case OPCODE_L:
	 filename_lookup = argv[j];
	 opcode = 0;
	 break;
       case OPCODE_F:
	 filename_word = argv[j];
	 opcode = 0;
	 break;
       case OPCODE_H:
	 hueristic = atoi(argv[j]);
	 opcode = 0;
	 break;
       case OPCODE_O:
	 output = atoi(argv[j]);
	 opcode = 0;
	 break;
      }
   }
   if ( (hueristic<0) || (hueristic>2) || (output<0) || (output>2)) {
      printf("Integer is not in valid range for argument.\n\n");
   }
   
   // 1.  We load the lookup file into a list.
   lookup_list = Tlookup_list_new ();  
   Tlookup_list_read_file (lookup_list, filename_lookup);
   
   // 2.  We load the input (text) file to be parsed as a list.
   word_list = Tword_list_new ();  
   Tword_list_read_file (word_list, filename_word);
   
   // 3.  We identify the pronouns and candidate proper nouns based on lookup and case.
   // 4.  We assign gender and number to all possible words based on lookup.
   current_word = Tlist_head (word_list);
   while (current_word != NULL) {
      current_lookup = Tlist_head (lookup_list);
      while (current_lookup != NULL) {
	 if (strcmp(current_lookup->string, current_word->string) == 0) {
	    if (current_lookup->type == PRONOUN) {
	       current_word->pronoun=1;
	       current_word->candidate= -1;  
	    }
	    if (current_lookup->type == ARTICLE) current_word->candidate=-1;
	    current_word->gender = current_lookup->gender;
	    current_word->number = current_lookup->number;
	    current_word->type = current_lookup->type;
	 }
	 current_lookup = Tlist_next_until_tail (lookup_list);
      }
      if ((current_word->string[0] < 'a') && (current_word->pronoun!=1)) 
	current_word->candidate = current_word->candidate + 1;
      else current_word->candidate = 0;
      current_word = Tlist_next_until_tail (word_list);
   }
   
   // 5.  We print the original text.  Either in plain text, tagged html, or the lispish format.
   switch (output) {
    case LISP:
      printf("#==> Lookup list structure\n");
      Tlookup_list_print_lispish (lookup_list);
      break;
    case TEXT:
      printf("\nThe original text follows.  \nPronouns are identified by an id in parenthesis.\n\n");
      Tword_list_print_plain (word_list);
      printf("\n");
      break;
    case HTML:
      printf("<html>\n"
	     "  <head>\n"
	     "    <title>Pronoun / Proper Noun Analysis</title>\n"
	     "  </head>\n"
	     "  <body bgcolor=\"ffffff\">\n<p>");
      printf("<h2>Pronoun / Proper Noun Analysis</h2>\n");
      printf("<h3>Marked up source text:</h3>\n");
      Tword_list_print_html (word_list);
      printf("<h3>Proper Noun Candidate(s) for Pronouns:</h3>\n");
      break;
   }
   
   // 6.  We set up a second (virtual) word list for candidates. 
   pronoun_list = Tword_list_new ();
   Tlist_virtual(pronoun_list, word_list);
   i=0;
   
   // 7.  We do some statistics on word relationships.
   //     (including gender and number checking).
   
   current_pronoun = Tlist_head (pronoun_list);
   while (current_pronoun != NULL) {
      if (current_pronoun->type == PRONOUN) {
	 switch (output) {
	  case HTML:
	    printf("<table cellpadding=\"2\" bgcolor=\"CCCCCC\"><tr><td>"
		   "<table cellpadding=\"4\" bgcolor=\"ffffff\">\n");
	    printf("    <tr>\n"
		   "      <td>Id</td>\n"
		   "      <td><a name=\"pronoun%d\">Pronoun</a></td>\n"
		   "      <td>Proper Noun Candidate</td>\n"
		   "      <td>Weight</td>\n"
		   "      <td>Gender</td>\n"
		   "      <td>Number</td>\n"
		   "    </tr>\n", i+1);
	    break;
	  case TEXT:
	    printf("Id  Pronoun  Proper Noun Candidate  Weight  Gender    Number  \n"
		   "==  =======  =====================  ======  ========  ========\n");
	    break;
	  case LISP:
	    printf("\n#==>  Candidate Weights\n");
	    printf("(Weights ");
	    break;
	 }

	 current_word = Tlist_head (word_list);
	 while (current_word != NULL) {
	    if (current_word->candidate == 1) {
	       int proximity_w, proximity_s, weight, gender, number;
	       // 8.  Push these proximity ratings onto lists.
	       proximity_w = current_pronoun->word - current_word->word;
	       proximity_s = current_pronoun->sentence - current_word->sentence;
	       Tword_list_sublist_push_back(current_word->proximity_w, proximity_w);
	       Tword_list_sublist_push_back(current_word->proximity_s, proximity_s);
	       // 9.  Check if the words are of the same gender.
	       if (current_pronoun->gender == current_word->gender) gender = 1;
	       else gender = 0;
	       Tword_list_sublist_push_back(current_word->gender_list, gender);
	       // 10.  Check if the words are of the same number.
	       if (current_pronoun->number == current_word->number) number = 1;
	       else number = 0;
	       Tword_list_sublist_push_back(current_word->number_list, number);
	       // 11. Create a hueristic weight based on these measures.
	       switch (hueristic) {
		case 0:
		case 1:
		case 2:
		  weight = (gender * 30)*number + (number * 30)*gender;
		  if (proximity_w != 0) weight = weight + (int)(20.0/(double)abs(proximity_w));
		  if (proximity_s != 0) weight = weight + (int)(5.0/(double)abs(proximity_s));
		  else weight = weight + 15;
		  break;
	       }
	       switch (output) {
		case HTML:
		  printf("    <tr>\n"
			 "      <td>%d</td>\n"
			 "      <td>%s</td>\n"
			 "      <td>%s</td>\n"
			 "      <td>%d</td>\n",
			 i+1, current_pronoun->string, current_word->string, weight);
		  if (current_word->gender == FEMALE) 
		    printf("      <td>Female</td>\n");
		  else if (current_word->gender == MALE) 
		    printf("      <td>Male</td>\n");
		  else 
		    printf("      <td>Unknown</td>\n");
		  if (current_word->number == PLURAL) 
		    printf("      <td>Plural</td>\n");
		  else if (current_word->number == SINGULAR) 
		    printf("      <td>Singular</td>\n");
		  else 
		    printf("      <td>Unknown</td>\n");
		  printf("    </tr>\n");
		  break;
		case TEXT:
		  printf("%2d  %7s  %21s  %3d     ", i+1, current_pronoun->string, 
			 current_word->string, weight);
		  if (current_word->gender == FEMALE) 
		    printf("Female     ");
		  else if (current_word->gender == MALE) 
		    printf("Male       ");
		  else 
		    printf("Unknown    ");
		  if (current_word->number == PLURAL) 
		    printf("Plural     ");
		  else if (current_word->number == SINGULAR) 
		    printf("Singular   ");
		  else 
		    printf("Unknown    ");
		  printf("\n");
		  break;
		case LISP:
		  printf("\n  (%d \"%s\" \"%s\" %d %d %d)", 
			 i+1, current_pronoun->string, 
			 current_word->string, weight, 
			 current_word->gender, current_word->number);
		  break;
	       }
	    }
	    current_word = Tlist_next_until_tail (word_list);
	 }
	 switch (output) {
	  case TEXT:
	    printf("\n");
	    break;
	  case HTML:
	    printf("  </table></tr></td></table>\n\n");
	    break;
	  case LISP:
	    printf(")\n");
	    break;
	 }
	 i=i+1;
      }
      current_pronoun = Tlist_next_until_tail (pronoun_list);
   }

   switch (output) {
    case LISP:
      printf("\n#==> Lookup list structure\n");
      Tword_list_print_lispish (word_list);
      break;
    case HTML:
      printf("  </body>\n"
	     "</html>\n");
      break;
    case TEXT:
      break;
   }
   Tlookup_list_delete (lookup_list);  
   Tword_list_delete (word_list);  
   return(0);   
}

